﻿local L = LibStub("AceLocale-3.0"):GetLocale("Critters", false)

Critters = CreateFrame("frame")
local Critters = Critters
Critters:SetScript("OnEvent", function(self, event, ...) if self[event] then return self[event](self, event, ...) end end)
Critters:RegisterEvent("ADDON_LOADED")
function Critters:Print(...) ChatFrame1:AddMessage(string.join(" ", "|cFF33FF99Critters|r:", ...)) end
local wipe = wipe
local CritterDBver = 1

-- create a tooltip scanner
local Scanner = CreateFrame("GameTooltip", "Critters_ScannerTooltip", nil, "GameTooltipTemplate")
Scanner:SetOwner(UIParent, "ANCHOR_NONE")
for i = 1, 30 do
	Scanner[i] = _G["Critters_ScannerTooltipTextLeft"..i]
	if not Scanner[i] then
		Scanner[i] = Scanner:CreateFontString()
		Scanner:AddFontStrings(Scanner[i], Scanner:CreateFontString())
	end
end

function Critters:ADDON_LOADED(event, addon)
	if addon ~= "Critters" then return end
	local _, class = UnitClass("player")
	for k, macro in pairs (self.macros[class]) do
		self.macros['GENERAL'][k] = macro
	end
	self.locked = {}
	for k in pairs (self.macros['GENERAL']) do
		self.locked[k] = true
	end
	defaults = {
		['class'] = {
			cm = 'Default',
		},
		['char'] = {
			cmloc = 'class',
			enablemacro = true,
			enableautosummon = true,
			cmenabled = false,
			clickreagents = true,
			autoreagents = false,
			Critters = {},
		},
		['global'] = {
			['macros'] = self.macros['GENERAL'],
		}
	}
	self.db = LibStub("AceDB-3.0"):New("CrittersDB", defaults, 'class')
	if self.db.char.CritterDBver ~= CritterDBver then
		self.db.char.CritterDBver = CritterDBver
		self.db.char.Critters = {}
	end
	self:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil
	if IsLoggedIn() then self:PLAYER_LOGIN() else self:RegisterEvent("PLAYER_LOGIN") end
	
	-- register SlashCommand
	SLASH_Critters1 = "/Critters";
	SlashCmdList["Critters"] = function()
		InterfaceOptionsFrame_OpenToCategory(self.optframe.Critters)
	end
end

function Critters:PLAYER_LOGIN()
	if GetMacroIndexByName("Critters") == 0 then
		if self.db.char.enablemacro then
			CreateMacro("Critters", 1, '')
		else
			DeleteMacro("Critters")
		end
	end
	self:Scan()
	if self.db.char.enablemacro then
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
	end
	if self.db.char.enableautosummon then
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("COMPANION_LEARNED")
		self:RegisterEvent("PLAYER_UNGHOST")
		self:RegisterEvent("PLAYER_CONTROL_GAINED")
		self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	end
	self:RegisterEvent("COMPANION_LEARNED")
	self.PLAYER_LOGIN = nil
	self:UpdateOptions()
end

function Critters:PLAYER_REGEN_DISABLED()
	self:UpdateMacro(true)
end

function Critters:PLAYER_REGEN_ENABLED()
	self:UpdateMacro()
end

function Critters:COMPANION_LEARNED()
	self:Scan()
end

function Critters:PLAYER_CONTROL_GAINED()
	self:CallCritter()
end

function Critters:PLAYER_ENTERING_WORLD()
	self:CallCritter()
end

function Critters:COMPANION_LEARNED()
	self:CallCritter()
end

function Critters:PLAYER_UNGHOST()
	self:CallCritter()
end

function Critters:ZONE_CHANGED_NEW_AREA()
	self:CallCritter()
end

function Critters:UNIT_SPELLCAST_SUCCEEDED(event, arg1, arg2)
	if arg1 == 'player' and self.db.char.Critters[self.ReverseLookup[arg2]] then
		self:UpdateMacro()
	end
end

function Critters:UpdateMacro(combat)
	if not InCombatLockdown() then
		if self.db.char.cmenabled and combat then
			EditMacro(GetMacroIndexByName("Critters"), "Critters", 1, self:ParseMacro(self.db.global.macros[self.db[self.db.char.cmloc].cm]))
		else
			local critter = self:RandomCritter(true)
			if critter then
				EditMacro(GetMacroIndexByName("Critters"), "Critters", 1, string.format('#showtooltip\n/cast %s()', GetSpellInfo(critter)))
			end
		end
	end
end

local blacklist = {["Throne of Kil'jaeden"] = true}
local lastcall = 0
function Critters:CallCritter()
	local currentcall = GetTime()
	if currentcall - lastcall < 10 then return end
	lastcall = currentcall
	local _, instanceType = IsInInstance()
	local pvping = instanceType == "pvp" or instanceType == "arena"
	if pvping or UnitIsPVP('player') or UnitIsPVPFreeForAll('player') or not HasFullControl() or IsStealthed() or UnitIsDeadOrGhost("player") or IsMounted() or IsFlying() or UnitCastingInfo("player") or blacklist[GetSubZoneText()] then
		return
	end
	local critter = self:RandomCritter(false)
	if critter then
		CallCompanion("CRITTER", self.SpellToID[critter])
	end
end

local bugs = {}
function Critters:RandomCritter(macro)
	wipe(bugs)
	for k, priority in pairs(self.db.char.Critters) do
		if priority ~= 0 then
			for i=1, priority do
				bugs[#bugs+1] = k
			end
		end
	end
	if not next(bugs) then return nil end
	local critter = bugs[math.random(1,#bugs)]
	local nr, hr = self:NeedsReagents(critter)
	if macro then
		if nr or hr and self.db.char.clickreagents then
			self:RandomCritter(true)
		else
			return critter
		end
	else
		if nr or hr and self.db.char.autoreagents then
			self:RandomCritter(false)
		else
			return critter
		end
	end
end

Critters.SpellToID = {}
Critters.ReverseLookup = {}
function Critters:Scan()
	wipe(self.SpellToID)
	local creatureSpellID
	for id = 1,GetNumCompanions("CRITTER") do
		creatureSpellID = select(3, GetCompanionInfo("CRITTER", id))
		self.SpellToID[creatureSpellID] = id
		self.ReverseLookup[GetSpellInfo(creatureSpellID)] = creatureSpellID
		self.db.char.Critters[creatureSpellID] = self.db.char.Critters[creatureSpellID] or 5
	end
	self:UpdateMacro()
	self:UpdateOptions()
end

function Critters:ParseMacro(macrotext)
	local spells = {}
	for mtype, mspell in string.gmatch(macrotext, "<(%a)(%w+)>") do
		if mtype == 's' then
			spells[#spells+1] = GetSpellInfo(mspell)
		end
	end
	return string.format(string.gsub(macrotext, '<%w+>', '%%s'), unpack(spells))
end

-- Returns true if Itemname is found in bags
function Critters:ScanBags(itemname)
	for bagID = 0, 4 do
		for slotID = 1, GetContainerNumSlots(bagID) do
			if GetContainerItemLink(bagID, slotID) and itemname == GetItemInfo(GetContainerItemLink(bagID, slotID)) then
				return true
			end
		end
	end
	return false
end

-- returns true if we need reagents
-- returns true, true if we have reagents
-- return false if we need nothing
local NeedsReagentsCache = {} -- do not cache critters that need reagets even if they are present
function Critters:NeedsReagents(critterspell)
	if NeedsReagentsCache[critterspell] then return false end
	Scanner:ClearLines()
	Scanner:SetHyperlink('spell:'..critterspell)
	for i=1,Scanner:NumLines() do
		if string.find (Scanner[i]:GetText(), SPELL_REAGENTS ,1, true) then
			if self:ScanBags(string.gsub(Scanner[i]:GetText(), SPELL_REAGENTS, '')) then
				return true, true
			else
				return true
			end
		end
	end
	NeedsReagentsCache[critterspell] = true
	return false
end
